# freeVikings - the "Lost Vikings" clone
# Copyright (C) 2005 Jakub "igneus" Pavlik

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


# viking.rb
# igneus 20.1.2004

require 'sprite.rb'
require 'deadviking.rb'

require 'vikingstate.rb'
require 'imagebank.rb'
require 'nullocation.rb'
require 'log4r'
require 'collisiontest.rb'

module FreeVikings

  class Viking < Sprite
    # Sprite trida pro postavicky vikingu

    BASE_VELOCITY = 65

    def initialize(name, start_position=[0,0])
      super()
      @log = Log4r::Logger['viking log']
      @name = name
      @state = VikingState.new
      @log.debug("Viking #{@name} initialised.")
      @rect = Rectangle.new start_position[0], start_position[1], 0, 0
      @last_update_time = Time.now.to_f
      @location = NullLocation.new # objekt overujici moznost presunu na posici
      @energy = 3 # zivotni sila

      @portrait = Portrait.new('viking_face.tga', 'viking_face_unactive.tga', 'dead_face.png')
    end

    def Viking.createWarior(name, start_position)
      return Warior.new(name, start_position)
    end

    def Viking.createSprinter(name, start_position)
      return Sprinter.new(name, start_position)
    end

    def Viking.createShielder(name, start_position)
      return Shielder.new(name, start_position)
    end

    attr_reader :name
    attr_reader :state
    attr_reader :energy
    attr_reader :portrait

    def paint(surface)
      surface.blit(@image.image(@state.to_s), coordinate_in_surface(surface))
    end

    def hurt
      @energy -= 1
      destroy if @energy <= 0
    end

    def move_left
      @state.move_left
      set_move
    end

    def move_right
      @state.move_right
      set_move
    end

    def stop
      @state.stop
    end

    def destroy
      @energy = 0
      @location.add_sprite DeadViking.new([left, top])
      @location.delete_sprite self
    end

    def alive?
      @energy > 0
    end

    def top
      @rect.top
    end

    def left
      @rect.left
    end

    def space_func_on
    end

    def space_func_off
    end

    def d_func_on
    end

    def d_func_off
    end

    # vrati souradnice stredu aktualniho obrazku

    def center
      [left + (@image.image.w / 2), top + (@image.image.w / 2)]
    end

    def moving?
      @state.moving?
    end

    def falling?
      @state.falling?
    end

    def standing?
      @state.standing?
    end

    def next_position
      time_now = Time.now.to_f
      time_delta = time_now - @last_update_time
      # Zde se musi posice zjistovat primo z instancni promenne, protoze
      # pristupove metody left a top ji aktualisuji
      next_top = @rect.top + (velocity_vertic * time_delta)
      next_left = @rect.left + (velocity_horiz * time_delta)
      Rectangle.new next_left, next_top, image.w, image.h
    end

    # Aktualisuje posici vikinga.

    def update
      # Nyni muzeme aktualisovat posici:
      if @location.is_position_valid?(self, next_position) then
	@log.debug "update: Viking #{name}'s next position is all right."
	@rect = next_position
	update_time
      else
	@log.debug "update: Viking #{name}'s next position isn't valid, he'll stuck now."
	@state.stop
        @state.descend
      end
      if @state.falling? and on_ground? then
        @state.descend
      end
      # Zkusme, jestli by viking nemohl zacit padat.
      # Pokud muze zacit padat, zacne padat:
      if not @state.rising? and not @state.falling? and not on_ground?
	@state.fall
	@log.debug "update: #{@name} starts falling because there's a free space under him."
      end

      @log.debug("update: #{@name}'s state: #{@state.to_s} #{@state.dump}")
    end

    private
    def velocity_vertic
      @state.velocity_vertic * BASE_VELOCITY
    end

    private
    def velocity_horiz
      @state.velocity_horiz * BASE_VELOCITY
    end

    private
    # Aktualisuje cas posledni aktualisace posice
    def update_time
      @last_update_time = Time.now.to_f
    end

    private
    def set_move
      update_time
    end

    private
    # Zjisti, jestli viking stoji na zemi (na pevne dlazdici)
    def on_ground?
      # stoji viking na stite?
      return true if on_shield?
      # je pod vikingem volne misto?
      lowerpos = next_position
      lowerpos[1] += 2
      return nil if @location.is_position_valid?(self, lowerpos)
      # viking stoji na pevne zemi:
      return true
    end

    private
    def on_shield?
      shield = @location.sprites_on_rect(rect).find {|s| s.is_a? Shield}
      if shield and CollisionTest.bottom_collision?(rect, shield.rect) then
        return true
      else
        return false
      end
    end

  end # class Viking
end # module

require 'warior.rb'
require 'sprinter.rb'
require 'shielder.rb'
